/**
 * Copyright (C) 2008-2012 AgroSense Foundation.
 *
 * AgroSense is free software: you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation, either version 3 of the License, or (at your option) any later
 * version.
 *
 * There are special exceptions to the terms and conditions of the GPLv3 as it
 * is applied to this software, see the FLOSS License Exception
 * <http://www.agrosense.eu/foss-exception.html>.
 *
 * AgroSense is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
 * A PARTICULAR PURPOSE. See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * AgroSense. If not, see <http://www.gnu.org/licenses/>.
 */
package nl.cloudfarming.client.sensor;

import eu.agrosense.client.grid.Grid;
import java.awt.event.ActionEvent;
import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.io.IOException;
import java.io.Serializable;
import java.util.List;
import javax.swing.AbstractAction;
import javax.swing.Action;
import javax.swing.Icon;
import nl.cloudfarming.client.geoviewer.BaseLayer;
import nl.cloudfarming.client.geoviewer.Layer;
import nl.cloudfarming.client.geoviewer.RangedLegendPalette;
import nl.cloudfarming.client.model.EntitySavable;
import nl.cloudfarming.client.sensor.model.Measurement;
import nl.cloudfarming.client.sensor.model.MeasurementCollection;
import nl.cloudfarming.client.sensor.model.Observation;
import nl.cloudfarming.client.sensor.model.ObservationManager;
import nl.cloudfarming.client.sensor.visual.util.ObservationUtil;
import nl.cloudfarming.client.util.UOM;
import org.netbeans.api.actions.Savable;
import org.netbeans.core.api.multiview.MultiViews;
import org.openide.awt.ActionID;
import org.openide.awt.ActionReference;
import org.openide.awt.ActionReferences;
import org.openide.filesystems.FileObject;
import org.openide.filesystems.MIMEResolver;
import org.openide.loaders.DataObject;
import org.openide.loaders.DataObjectExistsException;
import org.openide.loaders.MultiDataObject;
import org.openide.loaders.MultiFileLoader;
import org.openide.nodes.*;
import org.openide.util.ImageUtilities;
import org.openide.util.Lookup;
import org.openide.util.NbBundle;
import org.openide.util.lookup.AbstractLookup;
import org.openide.util.lookup.InstanceContent;
import org.openide.util.lookup.Lookups;
import org.openide.util.lookup.ProxyLookup;
import org.openide.windows.TopComponent;

/**
 * Data object for {@link Observation}
 *
 * @author Merijn Zengers <Merijn.Zengers@gmail.nl>
 */
@NbBundle.Messages({"ObservationDataObject data object name=Observation",
    "LBL_Observation_LOADER=Observation files"})
@MIMEResolver.ExtensionRegistration(displayName = "#LBL_Observation_LOADER",
mimeType = ObservationDataObject.MIME_TYPE,
extension = {"obs", "OBS"})
@DataObject.Registration(displayName = "#ObservationDataObject data object name", iconBase = ObservationDataObject.ICON_BASE, mimeType = ObservationDataObject.MIME_TYPE)
@ActionReferences(value = {
    @ActionReference(id =
    @ActionID(category = "System", id = "org.openide.actions.OpenAction"), path = "Loaders/application/x-agrosense-observation/Actions", position = 100, separatorAfter = 200),
    @ActionReference(id =
    @ActionID(category = "Edit", id = "org.openide.actions.DeleteAction"), path = "Loaders/application/x-agrosense-observation/Actions", position = 600),
    @ActionReference(id =
    @ActionID(category = "System", id = "org.openide.actions.ToolsAction"), path = "Loaders/application/x-agrosense-observation/Actions", position = 1300),})
public class ObservationDataObject extends MultiDataObject {

    public static final String MIME_TYPE = "application/x-agrosense-observation";
    public static final String ICON_BASE = "nl/cloudfarming/client/sensor/observation.png";
    private Observation observation;
    private final InstanceContent ic = new InstanceContent();
    private final Lookup lkp = new AbstractLookup(ic);
    private ObservationManager manager;
    private ObservationPropertyChangeListener changeListener = new ObservationPropertyChangeListener();

    public ObservationDataObject(FileObject fo, MultiFileLoader loader) throws DataObjectExistsException {
        super(fo, loader);
        registerEditor(MIME_TYPE, true);
        manager = Lookup.getDefault().lookup(ObservationManager.class);
        observation = manager.load(fo);
        assert observation != null : "Cannot load observation from primairy file";
        addListeners();
    }

    @Override
    protected Node createNodeDelegate() {
        ic.add(observation);
        return new ObservationNode(super.createNodeDelegate(), observation);
    }

    @Override
    protected int associateLookup() {
        return 1;
    }

    @Override
    public Lookup getLookup() {
        return new ProxyLookup(super.getLookup(), lkp);
    }

    /**
     * Handle deletion of the data object Removes savables from the lookup
     *
     * @throws IOException
     */
    @Override
    protected void handleDelete() throws IOException {
        manager.remove(observation);
        ObservationSavable savable = this.getLookup().lookup(ObservationSavable.class);
        if (null != savable) {
            savable.destroy();
        }
    }

    /**
     * Add property change listeners to the {@link Observation} and it's
     * {@link Measurement}'s
     */
    private void addListeners() {
        observation.addPropertyChangeListener(changeListener);
        for (MeasurementCollection collection : observation.getMeasurementCollections()) {
            collection.addPropertyChangeListener(MeasurementCollection.PROP_MEASUREMENTS, changeListener);
            for (Measurement measurement : collection.getMeasurements()) {
                measurement.addPropertyChangeListener(Measurement.PROP_VALUE, changeListener);
            }
        }
    }

    /**
     * Handles property change events of type
     * {@link MeasurementCollection#PROP_MEASUREMENTS} Will add listener if old
     * value is null and new value is not Will remove listener if old value is
     * not null and new value is null
     *
     * @param changeEvent
     */
    private void handleListeners(PropertyChangeEvent changeEvent) {
        switch (changeEvent.getPropertyName()) {
            case Observation.PROP_MEASUREMENT_COLLECTIONS:
                if (changeEvent.getOldValue() == null && changeEvent.getNewValue() != null) {
                    MeasurementCollection added = (MeasurementCollection) changeEvent.getNewValue();
                    added.addPropertyChangeListener(changeListener);
                    for (Measurement measurement : added.getMeasurements()) {
                        measurement.addPropertyChangeListener(Measurement.PROP_VALUE, changeListener);
                    }
                } else if (changeEvent.getOldValue() != null && changeEvent.getNewValue() == null) {
                    MeasurementCollection removed = (MeasurementCollection) changeEvent.getOldValue();
                    for (Measurement measurement : removed.getMeasurements()) {
                        measurement.removePropertyChangeListener(Measurement.PROP_VALUE, changeListener);
                    }
                    removed.removePropertyChangeListener(changeListener);
                }
                break;
            case MeasurementCollection.PROP_MEASUREMENTS:
                if (changeEvent.getOldValue() == null && changeEvent.getNewValue() != null) {
                    Measurement added = (Measurement) changeEvent.getNewValue();
                    added.addPropertyChangeListener(changeListener);
                } else if (changeEvent.getOldValue() != null && changeEvent.getNewValue() == null) {
                    Measurement removed = (Measurement) changeEvent.getOldValue();
                    removed.removePropertyChangeListener(changeListener);
                }
                break;
        }
    }

    /**
     * Listener for property changes of {@link PropertyChangeEvent#propertyName}
     * {@link MeasurementCollection#PROP_MEASUREMENTS} and
     * {@link MeasurementCollection#PROP_MEASUREMENT_TYPE} Will update the grid
     * if one of them changes. All other changes will result in a savable in the
     * lookup
     */
    private class ObservationPropertyChangeListener implements PropertyChangeListener {

        @Override
        public void propertyChange(PropertyChangeEvent changeEvent) {
            handleListeners(changeEvent);
            Savable savable =
                    getLookup().
                    lookup(Savable.class);
            if (null == savable) {
                ObservationDataObject.this.ic.add(new ObservationSavable());
            }
            switch (changeEvent.getPropertyName()) {
                case MeasurementCollection.PROP_MEASUREMENTS:
                    //Observation is added or removed the grid must be updated if there is any
                    //Only update grid if there is one
                    Measurement old = (Measurement) changeEvent.getOldValue();
                    Measurement newM = (Measurement) changeEvent.getNewValue();
                    MeasurementCollection collection = (MeasurementCollection) changeEvent.getSource();
                    Grid toUpdate = collection.getGrid();
                    if (toUpdate != null) {
                        List<Measurement> inGrid = collection.getMeasurements();
                        if (old != null && newM == null) {
                            //removed.
                            ObservationUtil.removeMeasurementFromGrid(inGrid, toUpdate, old);
                        } else if (old == null && newM != null) {
                            //added
                            ObservationUtil.addMeasurementToGrid(inGrid, toUpdate, newM);
                        }
                    }
                    break;
            }
        }
    }
    private static final Icon ICON = ImageUtilities.loadImageIcon("nl/cloudfarming/client/sensor/observation.png", true);

    /**
     * Savable for observations
     */
    private class ObservationSavable extends EntitySavable {

        public ObservationSavable() {
            super(observation, ICON);
            register();
        }

        @Override
        public void handleSave() throws IOException {
            dataObject().manager.save(observation);
            destroy();
        }

        @Override
        protected String findDisplayName() {
            return observation.getName();
        }

        @Override
        public void destroy() {
            dataObject().ic.remove(ObservationSavable.this);
            unregister();
        }

        @Override
        public void dispose() {
            dataObject().dispose();
        }

        ObservationDataObject dataObject() {
            return ObservationDataObject.this;
        }

        @Override
        public boolean equals(Object obj) {
            if (obj instanceof ObservationSavable) {
                ObservationSavable m = (ObservationSavable) obj;
                return dataObject() == m.dataObject();
            }
            return false;
        }

        @Override
        public int hashCode() {
            return dataObject().hashCode();
        }
    }

    /**
     * Node which can display a {@link Observation}
     */
    private class ObservationNode extends FilterNode {

        private Observation observation;
        private Node original;

        public ObservationNode(Node original, Observation observation) {
            super(original, Children.create(new ObservationChildFactory(observation), true), original.getLookup());
            this.observation = observation;
            this.original = original;
            observation.addPropertyChangeListener(new PropertyChangeListener() {
                @Override
                public void propertyChange(PropertyChangeEvent evt) {
                    if (Observation.PROP_NAME.equals(evt.getPropertyName()) || Observation.PROP_SAMPLE_DATE.equals(evt.getPropertyName())) {
                        setDisplayName(getDisplayName());
                    }
                }
            });
        }

        @Override
        public String getDisplayName() {
            StringBuilder nameBuilder = new StringBuilder(observation.getName());
            nameBuilder.append(" (");
            nameBuilder.append(observation.getSampleDate());
            nameBuilder.append(")");
            return nameBuilder.toString();
        }

        @Override
        public Action[] getActions(boolean context) {
            return original.getActions(context);
        }
    }

    /**
     * Child factory which produces {@link MeasurementNode}'s for
     * {@link Measurement}'s
     */
    private class ObservationChildFactory extends ChildFactory<MeasurementCollection> {

        private final List<MeasurementCollection> collections;
        private MeasurementListener measurementListener = new MeasurementListener();

        public ObservationChildFactory(Observation observation) {
            this.collections = observation.getMeasurementCollections();
            observation.addPropertyChangeListener(Observation.PROP_MEASUREMENT_COLLECTIONS, measurementListener);
        }

        @Override
        protected boolean createKeys(List<MeasurementCollection> toPopulate) {
            toPopulate.addAll(collections);
            return true;
        }

        @Override
        protected Node createNodeForKey(MeasurementCollection key) {
            if (key.getGrid() == null) {
                Grid grid = ObservationUtil.createGrid(observation, key.getMeasurements());
                key.setGrid(grid);
            }
            return new MeasurementNode(key);
        }

        private void refresh() {
            refresh(true);
        }

        private class MeasurementListener implements PropertyChangeListener {

            @Override
            public void propertyChange(PropertyChangeEvent evt) {
                switch (evt.getPropertyName()) {
                    case Observation.PROP_MEASUREMENT_COLLECTIONS:
                        refresh();
                        break;
                }
            }
        }
    }

    /**
     * Node which displays properties of a {@link Measurement}
     */
    private static class MeasurementNode extends AbstractNode implements Serializable, Lookup.Provider {

        private MeasurementCollection collection;

        public MeasurementNode(MeasurementCollection collection) {
            super(Children.LEAF, Lookups.fixed(collection, collection.getGrid(), new BaseLayer(getPalette(collection), collection.getMeasurementType().toString())));
            this.collection = collection;

            setName(collection.getMeasurementType().toString());
            setDisplayName(collection.getMeasurementType().toString());
            setIconBaseWithExtension("nl/cloudfarming/client/sensor/measurement.png");

            PaletteUpdater paletteUpdater = new PaletteUpdater();
            collection.addPropertyChangeListener(MeasurementCollection.PROP_MEASUREMENTS, paletteUpdater);
            for (Measurement measurement : collection.getMeasurements()) {
                measurement.addPropertyChangeListener(Measurement.PROP_VALUE, paletteUpdater);
            }
        }

        @Override
        public Action[] getActions(boolean arg0) {
            Action[] nodeActions = new Action[1];
            nodeActions[0] = new OpenMeasurementDetailsAction();
            return nodeActions;
        }

        private static RangedLegendPalette getPalette(MeasurementCollection collection) {
            Double minValue = Double.MAX_VALUE;
            Double maxValue = Double.MIN_VALUE;
            for (Measurement measurement : collection.getMeasurements()) {
                if (measurement.getValue() < minValue) {
                    minValue = measurement.getValue();
                }
                if (measurement.getValue() > maxValue) {
                    maxValue = measurement.getValue();
                }
            }
            return new RangedLegendPalette<>(minValue, maxValue, collection.getMeasurementType().getUnitOfMeasure());
        }

        private void updatePalette() {
            Double minValue = Double.MAX_VALUE;
            Double maxValue = Double.MIN_VALUE;
            for (Measurement measurement : collection.getMeasurements()) {
                if (measurement.getValue() < minValue) {
                    minValue = measurement.getValue();
                }
                if (measurement.getValue() > maxValue) {
                    maxValue = measurement.getValue();
                }
            }
            RangedLegendPalette palette = (RangedLegendPalette) getLookup().lookup(Layer.class).getPalette();
            palette.setMinMaxValue(minValue, maxValue);
        }

        private class PaletteUpdater implements PropertyChangeListener {

            @Override
            public void propertyChange(PropertyChangeEvent evt) {
                updatePalette();
            }
        }

        @NbBundle.Messages(value = "OpenMeasurementDetailsAction action name=Open")
        private final class OpenMeasurementDetailsAction extends AbstractAction {

            public OpenMeasurementDetailsAction() {
                super(Bundle.OpenMeasurementDetailsAction_action_name());
            }

            @Override
            public void actionPerformed(ActionEvent e) {
                TopComponent tc = MultiViews.createMultiView(Measurement.MIME_TYPE, MeasurementNode.this);
                tc.open();
                tc.requestActive();
            }
        }
    }
}
