/**
 * Copyright (C) 2008-2012 AgroSense Foundation.
 *
 * AgroSense is free software: you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation, either version 3 of the License, or (at your option) any later
 * version.
 *
 * There are special exceptions to the terms and conditions of the GPLv3 as it
 * is applied to this software, see the FLOSS License Exception
 * <http://www.agrosense.eu/foss-exception.html>.
 *
 * AgroSense is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
 * A PARTICULAR PURPOSE. See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * AgroSense. If not, see <http://www.gnu.org/licenses/>.
 */
package nl.cloudfarming.client.sensor;

import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.OutputStream;
import java.util.*;
import java.util.logging.Level;
import java.util.logging.Logger;
import nl.cloudfarming.client.farm.model.FarmManager;
import nl.cloudfarming.client.model.AgroURI;
import nl.cloudfarming.client.model.ItemIdType;
import nl.cloudfarming.client.sensor.model.Observation;
import nl.cloudfarming.client.sensor.model.ObservationManager;
import org.netbeans.api.project.Project;
import org.openide.filesystems.FileObject;
import org.openide.filesystems.FileUtil;
import org.openide.util.Exceptions;
import org.openide.util.Lookup;
import org.openide.util.lookup.ServiceProvider;

/**
 * Manager which can load and save a {@link Observation}
 *
 * @author Merijn Zengers <Merijn.Zengers@gmail.nl>
 */
@ServiceProvider(service = ObservationManager.class)
public class ObservationManagerImpl extends ObservationManager {

    private static final Set<ItemIdType> SUPPORTED_TYPES = EnumSet.of(ItemIdType.OBS);
    private static final Logger LOGGER = Logger.getLogger("nl.cloudfarming.client.sensor.ObservationManagerImpl");
    private Map<AgroURI, Set<AgroURI>> index = new HashMap<>();
    private static final String INDEX_FOLDER = "ObservationIndex";
    private static final String INDEX_FILE = "observationIndex.idx";

    /**
     * Load a Observation by it's uri
     *
     * @param uri
     * @param project
     * @return
     */
    @Override
    protected Observation load(AgroURI uri, Project project) {
        FileObject observationObject = getObservationsFolder(project, true).getFileObject(uri.getItemIdNumber(), Observation.EXT);
        return load(observationObject);
    }

    /**
     * Loads an observation for a specified {@link FileObject}
     *
     * @param observationObject
     * @return
     */
    @Override
    public Observation load(FileObject observationObject) {
        LOGGER.log(Level.FINEST, "loading Observation object from file {0}", observationObject.getName());
        assert Observation.EXT.equals(observationObject.getExt()) : "Can only load Observations";
        Object object = null;

        try (ObjectInputStream ois = new ObjectInputStream(observationObject.getInputStream())) {
            object = ois.readObject();
        } catch (ClassNotFoundException | IOException ex) {
            LOGGER.log(Level.SEVERE, "Could not load file {0} with the exception {1}", new Object[]{observationObject, ex});
        }
        assert object instanceof Observation : "fileobject did not contain the right type (Observation). Loaded type is " + object;
        Observation loaded = (Observation) object;
        //Add to index
        addToIndex(loaded.getSource(), loaded.getURI());
        return getCachedInstanceFor(loaded);
    }

    /**
     * Save a observation
     *
     * @param entity the entity to save
     */
    @Override
    public void save(Observation entity) {
        LOGGER.log(Level.FINEST, "Saving Observation with id {0}", new Object[]{entity.getItemIdNumber()});
        FarmManager farmManager = Lookup.getDefault().lookup(FarmManager.class);
        Project project = farmManager.findFarmProject(entity.getURI());

        FileObject observationsFolder = getObservationsFolder(project, true);

        FileObject observationObject = observationsFolder.getFileObject(entity.getItemIdNumber(), Observation.EXT);

        try {
            if (observationObject == null) {
                String fileName = entity.getItemIdNumber() + "." + Observation.EXT;
                try (OutputStream os = observationsFolder.createAndOpen(fileName)) {
                    save(entity, os, fileName);
                    //Added property change
                    getPropertyChangeSupport().firePropertyChange(PROP_OBSERVATION, null, entity);
                }
            } else {
                save(entity, observationObject);
            }

        } catch (IOException ex) {
            LOGGER.log(Level.SEVERE, "Could not create Observation file {0} with the exception {1}", new Object[]{entity.getItemIdNumber(), ex});
        }
    }

    /**
     * Save an Observation with the output stream and filename present
     *
     * @param entity
     * @param outputStream
     * @param fileName
     */
    protected void save(Observation entity, OutputStream outputStream, String fileName) {
        try (ObjectOutputStream oos = new ObjectOutputStream(outputStream)) {
            oos.writeObject(entity);
        } catch (IOException ex) {
            LOGGER.log(Level.SEVERE, "Could not save file {0} with the exception {1}", new Object[]{fileName, ex});
        }
        addToCache(entity);
        addToIndex(entity.getSource(), entity.getURI());
    }

    /**
     * Save an Observation to a specific file object
     *
     * @param object
     * @param fileObject
     */
    protected void save(Observation object, FileObject fileObject) {
        try {
            save(object, fileObject.getOutputStream(), fileObject.getNameExt());
        } catch (IOException ex) {
            LOGGER.log(Level.SEVERE, "Could not save file {0} with the exception {1}", new Object[]{fileObject, ex});
        }
    }

    /**
     * Removes the entity Also removes from index
     *
     * @param entity
     */
    @Override
    public void remove(Observation entity) {
        FarmManager farmManager = Lookup.getDefault().lookup(FarmManager.class);
        Project project = farmManager.findFarmProject(entity.getURI());
        remove(entity, project);
    }

    /**
     * Removes a observation from a certain project also removes it from the
     * index
     *
     * @param entity
     * @param project
     */
    protected void remove(Observation entity, Project project) {
        FileObject observationObject = null;
        try {
            FileObject observationsFolder = getObservationsFolder(project, true);

            observationObject = observationsFolder.getFileObject(entity.getItemIdNumber(), Observation.EXT);

            observationObject.delete();
            removeFromIndex(entity.getSource(), entity.getURI());
            getPropertyChangeSupport().firePropertyChange(PROP_OBSERVATION, entity, null);
        } catch (IOException ex) {
            LOGGER.log(Level.SEVERE, "Could not delete file {0} with the exception {1}", new Object[]{observationObject, ex});
        }
    }

    /**
     * Get the supported ItemIdType
     *
     * @return
     */
    @Override
    public Set<ItemIdType> getSupportedTypes() {
        return SUPPORTED_TYPES;
    }

    /**
     * Adds a Observation agroUri to the index of the specified source agroURI
     *
     * @param source
     * @param fileName
     */
    protected void addToIndex(AgroURI source, AgroURI observationReference) {
        Set<AgroURI> references = index.get(source);
        if (references == null) {
            references = new HashSet<>();
            references.add(observationReference);
            index.put(source, references);
        } else {
            references.add(observationReference);
        }
    }

    /**
     * Remove a observation from the index
     *
     * @param source
     * @param observationReference
     */
    protected void removeFromIndex(AgroURI source, AgroURI observationReference) {
        Set<AgroURI> references = index.get(source);
        if (references != null) {
            references.remove(observationReference);
        }
    }

    /**
     * Get the indexed file names from the index for the provided {@link AgroURI}
     *
     * @param agroURI
     * @return
     */
    protected Set<AgroURI> getIndexedObservations(AgroURI agroURI) {
        return index.get(agroURI);
    }

    /**
     * Restore the index from the file it is saved in
     */
    void restoreIndex() {
        FileObject indexFileObject = getIndexFile(false);
        Object object = null;
        if (indexFileObject != null) {
            try (ObjectInputStream oos = new ObjectInputStream(indexFileObject.getInputStream())) {
                object = oos.readObject();
            } catch (IOException | ClassNotFoundException ex) {
                LOGGER.log(Level.SEVERE, "Could not load index file {0} with the exception {1}", new Object[]{indexFileObject.getName(), ex});
            }
            if (object instanceof Map) {
                index = (Map<AgroURI, Set<AgroURI>>) object;
            }
        }
        else{
            index = new HashMap<>();
        }
    }
    
    /**
     * Save the index to a file
     */
    void saveIndex(){
        FileObject indexFileObject = getIndexFile(true);
        try (ObjectOutputStream oos = new ObjectOutputStream(indexFileObject.getOutputStream())) {
            oos.writeObject(index);
        } catch (IOException ex) {
            LOGGER.log(Level.SEVERE, "Could not save index file {0} with the exception {1}", new Object[]{indexFileObject.getName(), ex});
        }
    }

    private FileObject getIndexFile(boolean create) {
        FileObject configRoot = FileUtil.getConfigRoot();
        FileObject indexFolder = configRoot.getFileObject(INDEX_FOLDER);
        if (indexFolder == null) {
            try {
                indexFolder = configRoot.createFolder(INDEX_FOLDER);
            } catch (IOException ex) {
                LOGGER.log(Level.WARNING, "Unable to create observation index folder in config root {0}", ex.getMessage());
            }
        }
        FileObject indexFileObject = indexFolder.getFileObject(INDEX_FILE);
        if (indexFileObject == null && create) {
            try {
                indexFileObject = indexFolder.createData(INDEX_FILE);
            } catch (IOException ex) {
                LOGGER.log(Level.WARNING, "Unable to create observation index file in config root {0}", ex.getMessage());
            }
        }
        return indexFileObject;
    }

    /**
     * Get observation file objects for a specific source uri
     */
    @Override
    public List<FileObject> getObservationFileObjectsForSource(AgroURI source) {
        if (source == null) {
            throw new IllegalArgumentException("Observations cannot be loaded from null source");
        }
        List<FileObject> observations = new ArrayList<>();
        FarmManager farmManager = Lookup.getDefault().lookup(FarmManager.class);
        Project project = farmManager.findFarmProject(source);
        FileObject observationsFolder = getObservationsFolder(project, true);
        //Get from index
        Set<AgroURI> indexedObservations = getIndexedObservations(source);
        if (indexedObservations != null) {
            for (AgroURI indexed : indexedObservations) {
                Observation loaded = load(indexed);
                if (source.equals(loaded.getSource())) {
                    FileObject observationObject = observationsFolder.getFileObject(loaded.getItemIdNumber(), Observation.EXT);
                    observations.add(observationObject);
                }
            }
        }
        return observations;
    }
}
