/**
 * Copyright (C) 2008-2012 AgroSense Foundation.
 *
 * AgroSense is free software: you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation, either version 3 of the License, or (at your option) any later
 * version.
 *
 * There are special exceptions to the terms and conditions of the GPLv3 as it
 * is applied to this software, see the FLOSS License Exception
 * <http://www.agrosense.eu/foss-exception.html>.
 *
 * AgroSense is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
 * A PARTICULAR PURPOSE. See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * AgroSense. If not, see <http://www.gnu.org/licenses/>.
 */
package nl.cloudfarming.client.sensor;

import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.util.List;
import javax.swing.Action;
import nl.cloudfarming.client.geoviewer.SimpleLayer;
import nl.cloudfarming.client.model.AgroURI;
import nl.cloudfarming.client.model.Entity;
import nl.cloudfarming.client.sensor.model.Observation;
import nl.cloudfarming.client.sensor.model.ObservationManager;
import org.openide.filesystems.FileObject;
import org.openide.loaders.DataObject;
import org.openide.loaders.DataObjectNotFoundException;
import org.openide.nodes.AbstractNode;
import org.openide.nodes.ChildFactory;
import org.openide.nodes.Children;
import org.openide.nodes.Node;
import org.openide.util.Exceptions;
import org.openide.util.Lookup;
import org.openide.util.NbBundle;
import org.openide.util.Utilities;
import org.openide.util.lookup.Lookups;
import org.openide.util.lookup.ProxyLookup;

/**
 * Provider for Observation nodes for a link {@link AgroURI}
 *
 * @author Merijn Zengers <Merijn.Zengers@gmail.nl>
 */
public class ObservationNodeProvider {

    /**
     * The action path of the observation node. If you want to register a action on this node use this path
     */
    public static final String ACTION_PATH = "Loaders/application/x-agrosense-observations/Actions";
    
    /**
     * Get the Observations node for a specific source
     * The returned node will have the provided entity in it's lookup
     *
     * @param source
     * @return
     */
    public static Node getObservationsNode(Entity source) {
        return new ObservationsRootNode(source);
    }
    

    /**
     * Root node for observations
     */
    @NbBundle.Messages({"ObservationsRootNode node display name=Observations"})
    private static class ObservationsRootNode extends AbstractNode {

        private static final String NAME = Bundle.ObservationsRootNode_node_display_name();
        

        public ObservationsRootNode(Entity source) {
            super(Children.create(new ObservationsChildFactory(source.getURI()), true), new ProxyLookup(Lookups.fixed(source), Lookups.fixed(new SimpleLayer())));
            setName(NAME);
            setDisplayName(NAME);
            setIconBaseWithExtension("nl/cloudfarming/client/sensor/observations.png");
        }

        @Override
        public Action[] getActions(boolean context) {
            List<? extends Action> myActions =
                    Utilities.actionsForPath(ACTION_PATH);
            return myActions.toArray(new Action[myActions.size()]);
        }
    }

    /**
     * Child factory which can create nodes for {@link Observation}'s
     */
    private static class ObservationsChildFactory extends ChildFactory<Node> {

        private ObservationManager manager;
        private AgroURI source;

        public ObservationsChildFactory(AgroURI source) {
            this.source = source;
            manager = Lookup.getDefault().lookup(ObservationManager.class);
            manager.addPropertyChangeListener(ObservationManager.PROP_OBSERVATION, new PropertyChangeListener() {

                @Override
                public void propertyChange(PropertyChangeEvent evt) {
                    refresh(true);
                }
            });
        }

        @Override
        protected boolean createKeys(List<Node> toPopulate) {
            List<FileObject> observations = manager.getObservationFileObjectsForSource(source);
            for (FileObject observationObject : observations) {
                try {
                    DataObject observationDataObject = (ObservationDataObject) DataObject.find(observationObject);
                    Node observationNode = observationDataObject.getNodeDelegate();
                    toPopulate.add(observationNode);
                } catch (DataObjectNotFoundException ex) {
                    Exceptions.printStackTrace(ex);
                }
            }
            return true;
        }

        @Override
        protected Node createNodeForKey(Node key) {
            return key;
        }
    }
}
