/**
 * Copyright (C) 2008-2012 AgroSense Foundation.
 *
 * AgroSense is free software: you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation, either version 3 of the License, or (at your option) any later
 * version.
 *
 * There are special exceptions to the terms and conditions of the GPLv3 as it
 * is applied to this software, see the FLOSS License Exception
 * <http://www.agrosense.eu/foss-exception.html>.
 *
 * AgroSense is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
 * A PARTICULAR PURPOSE. See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * AgroSense. If not, see <http://www.gnu.org/licenses/>.
 */
package nl.cloudfarming.client.sensor;

import eu.agrosense.client.grid.Grid;
import java.awt.Image;
import java.awt.datatransfer.Transferable;
import java.beans.PropertyChangeEvent;
import java.io.IOException;
import java.util.Arrays;
import javax.swing.Action;
import org.openide.nodes.FilterNode;
import org.openide.nodes.Node;
import org.openide.nodes.NodeTransfer;
import org.openide.util.ImageUtilities;
import org.openide.util.NbBundle;
import org.openide.util.datatransfer.PasteType;

/**
 * Node which represents sensor files
 *
 * @author Merijn Zengers
 */
@NbBundle.Messages({"SensorFileNode no data message=(no data found)"})
public abstract class SensorFileNode extends FilterNode {

    private Node original;

    /**
     * <p> Creates new SensorFile node Will place the provided layer and the
     * {@link SensorGrid} returned by {@link SensorDataObject#getSensorGrid()}
     * in it's lookup. </p> <p>adds {@link PaletteChangeListener} on provided
     * {@link SensorDataObject} to listen to {@link PropertyChangeEvent} of
     * {@link PropertyChangeEvent#propertyName} {@link SensorDataObject#PROP_PALETTE_CHANGED}
     * </p>
     *
     * @param original
     * @param sensorDataObject
     * @param layer
     */
    public SensorFileNode(Node original) {
        super(original, FilterNode.Children.LEAF, original.getLookup());
        this.original = original;
    }

    /**
     * Adds a {@link ChangePaletteAction} to the actions from the original node
     *
     * @param arg0
     * @return
     */
    @Override
    public Action[] getActions(boolean arg0) {
        Action[] addedActions = {new ChangePaletteAction()};

        Action[] originalActions = original.getActions(arg0);
        originalActions[0].setEnabled(true);
        Action[] actions = Arrays.copyOf(originalActions, originalActions.length + addedActions.length);

        int counter = originalActions.length;
        for (Action addedAction : addedActions) {
            actions[counter] = addedAction;
        }
        return actions;
    }

    @Override
    public PasteType getDropType(Transferable t, int action, int index) {
        // rely on AbstractNode behavior which checks if the transferable contains a NodeTransfer.Paste.
        return super.getDropType(t, action, index);
    }

    @Override
    public Transferable clipboardCopy() throws IOException {
        return NodeTransfer.transferable(this, NodeTransfer.CLIPBOARD_COPY);
    }

    @Override
    public Transferable clipboardCut() throws IOException {
        return clipboardCopy();
    }

    /**
     * This implementation only calls clipboardCopy supposing that copy to
     * clipboard and copy by d'n'd are similar.
     *
     * @return transferable to represent this node during a drag
     * @exception IOException when the cut cannot be performed
     */
    @Override
    public Transferable drag() throws IOException {
        return clipboardCopy();
    }

    public abstract Image getIcon();

    @Override
    public String getDisplayName() {
        Grid grid = getLookup().lookup(Grid.class);
        String displayName = original.getDisplayName();
        if (grid == null) {
            StringBuilder errorBuilder = new StringBuilder(displayName);
            errorBuilder.append(" ");
            errorBuilder.append(Bundle.SensorFileNode_no_data_message());
            return errorBuilder.toString();
        }
        return displayName;
    }

    /**
     * Searches for a {@link Grid} in the lookup if not found will add
     * exclamation to the icon
     *
     * @param type
     * @return
     */
    @Override
    public Image getIcon(int type) {
        Grid grid = getLookup().lookup(Grid.class);
        if (grid == null) {
            return ImageUtilities.mergeImages(getIcon(), ImageUtilities.loadImage("nl/cloudfarming/client/sensor/exclamation.png"), 0, 0);
        }
        return getIcon();
    }
}