/**
 * Copyright (C) 2008-2012 AgroSense Foundation.
 *
 * AgroSense is free software: you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation, either version 3 of the License, or (at your option) any later
 * version.
 *
 * There are special exceptions to the terms and conditions of the GPLv3 as it
 * is applied to this software, see the FLOSS License Exception
 * <http://www.agrosense.eu/foss-exception.html>.
 *
 * AgroSense is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
 * A PARTICULAR PURPOSE. See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * AgroSense. If not, see <http://www.gnu.org/licenses/>.
 */
package nl.cloudfarming.client.sensor.visual;

import com.vividsolutions.jts.geom.Coordinate;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.beans.PropertyChangeSupport;
import java.util.*;
import javax.swing.JTable;
import javax.swing.event.DocumentEvent;
import javax.swing.event.DocumentListener;
import javax.swing.event.TableModelEvent;
import javax.swing.event.TableModelListener;
import javax.swing.table.AbstractTableModel;
import nl.cloudfarming.client.sensor.model.Measurement;
import nl.cloudfarming.client.sensor.model.MeasurementCollection;
import nl.cloudfarming.client.sensor.model.MeasurementType;
import nl.cloudfarming.client.sensor.model.Observation;
import nl.cloudfarming.client.sensor.visual.util.ObservationUtil;
import nl.cloudfarming.client.util.UOM;
import nl.cloudfarming.client.util.swing.beanbinding.DateInPastValidator;
import nl.cloudfarming.client.util.swing.beanbinding.NonEmptyStringValidator;
import org.jdesktop.beansbinding.*;
import org.openide.util.NbBundle;

/**
 * Abstract controller for {@link ObservationVisualPanel}
 *
 * @author Merijn Zengers <Merijn.Zengers@gmail.nl>
 */
@NbBundle.Messages({"ObservationVisualController error message empty name=Name must be entered",
    "ObservationVisualController error message empty sample date=Date must be entered",
    "ObservationVisualController error message date in past=Date cannot be in the future",
    "ObservationVisualController error message empty measurement type=Type must be entered",
    "ObservationVisualController error message empty measurement value=Value must be entered"})
public abstract class ObservationVisualController {

    private BeanProperty formPropName = BeanProperty.create("text");
    private BeanProperty soilSamplePropName = BeanProperty.create(Observation.PROP_NAME);
    private BeanProperty formPropSampleDate = BeanProperty.create("date");
    private BeanProperty soilSamplePropSampleDate = BeanProperty.create(Observation.PROP_SAMPLE_DATE);
    private BindingGroup soilSampleBindingGroup = new BindingGroup();
    private ObservationVisualPanel visualComponent;
    private VisaulPanelChangeListener changeListener = new VisaulPanelChangeListener();
    private Observation observation;
    private MeasurementTableBinder measurementTableBinder;

    public ObservationVisualController(Observation observation) {
        this.observation = observation;
    }

    /**
     * Get the observation this controller works with
     *
     * @return
     */
    Observation getObservation() {
        return observation;
    }

    protected ObservationVisualPanel getPanel() {
        if (visualComponent == null) {
            visualComponent = new ObservationVisualPanel();
            bind();
            ButtonListener buttonListener = new ButtonListener();
            visualComponent.getAddButton().addActionListener(buttonListener);
            visualComponent.getRemoveButton().addActionListener(buttonListener);

            visualComponent.getNameField().getDocument().addDocumentListener(changeListener);
            visualComponent.getSampleDatePicker().addActionListener(changeListener);
            visualComponent.getMeasurementsTable().getModel().addTableModelListener(changeListener);
        }
        return visualComponent;
    }

    private void bind() {
        Binding nameBinding = Bindings.createAutoBinding(AutoBinding.UpdateStrategy.READ_WRITE,
                observation, soilSamplePropName, visualComponent.getNameField(), formPropName);
        nameBinding.setValidator(new NonEmptyStringValidator());
        soilSampleBindingGroup.addBinding(nameBinding);
        Binding sampleDateBinding = Bindings.createAutoBinding(AutoBinding.UpdateStrategy.READ_WRITE,
                observation, soilSamplePropSampleDate, visualComponent.getSampleDatePicker(), formPropSampleDate);
        sampleDateBinding.setValidator(new DateInPastValidator());
        soilSampleBindingGroup.addBinding(sampleDateBinding);
        soilSampleBindingGroup.bind();
        measurementTableBinder = new MeasurementTableBinder(observation, visualComponent.getMeasurementsTable());
    }

    /**
     * Check if the name is valid Only valid if not empty
     *
     * @return
     */
    private boolean isNameValid() {
        return !visualComponent.getNameField().getText().isEmpty();
    }

    /**
     * Check if the date is valid Only valid if not null
     *
     * @return
     */
    private boolean isDateValid() {
        return isDateFilled() && isDateInPast();
    }

    private boolean isDateFilled() {
        return visualComponent.getSampleDatePicker().getDate() != null;
    }

    private boolean isDateInPast() {
        Date now = new Date();
        return visualComponent.getSampleDatePicker().getDate().getTime() <= now.getTime();
    }

    /**
     * Check if the panel is valid
     *
     * @return
     */
    public boolean isPanelValid() {
        boolean valid = true;
        if (!isNameValid()) {
            valid = false;
        }
        if (!isDateValid()) {
            valid = false;

        }
        if (!measurementTableBinder.measurmentsValid()) {
            valid = false;
        }
        return valid;
    }

    /**
     * Validates the panel will add error notifications if not Use
     * {@link ObservationVisualController#isValid()} to check if the panel is
     * valid
     */
    public void validatePanel() {
        visualComponent.getNameRow().setValid();
        visualComponent.getSampleDateRow().setValid();
        visualComponent.getErrorPanel().setValid();
        if (!isNameValid()) {
            visualComponent.getNameRow().setInvalid(Bundle.ObservationVisualController_error_message_empty_name());
        }
        if (!isDateFilled()) {
            visualComponent.getSampleDateRow().setInvalid(Bundle.ObservationVisualController_error_message_empty_sample_date());
        } else if (!isDateInPast()) {
            visualComponent.getSampleDateRow().setInvalid(Bundle.ObservationVisualController_error_message_date_in_past());
        }

        if (!measurementTableBinder.measurementTypesValid()) {
            visualComponent.getErrorPanel().setError(Bundle.ObservationVisualController_error_message_empty_measurement_type());
        } else if (!measurementTableBinder.measurementsValueValid()) {
            visualComponent.getErrorPanel().setError(Bundle.ObservationVisualController_error_message_empty_measurement_value());
        }

    }

    /**
     * Fired when there are changes in the panel
     */
    public abstract void panelChanged();

    /**
     * Listens to the add and remove buttons
     */
    private class ButtonListener implements ActionListener {

        @Override
        public void actionPerformed(ActionEvent e) {
            switch (e.getActionCommand()) {
                case ObservationVisualPanel.ADD_BUTTON_ACTION_COMMAND:
                    measurementTableBinder.addRow();
                    panelChanged();
                    break;
                case ObservationVisualPanel.DELETE_BUTTON_ACTION_COMMAND:
                    for (int rowIndex : visualComponent.getMeasurementsTable().getSelectedRows()) {
                        measurementTableBinder.removeRow(rowIndex);
                    }
                    panelChanged();
                    break;
            }
        }
    }

    /**
     * Listens to changes in the panel
     */
    private class VisaulPanelChangeListener implements ActionListener, DocumentListener, TableModelListener {

        @Override
        public void actionPerformed(ActionEvent e) {
            panelChanged();
        }

        @Override
        public void insertUpdate(DocumentEvent e) {
            panelChanged();
        }

        @Override
        public void removeUpdate(DocumentEvent e) {
            panelChanged();
        }

        @Override
        public void changedUpdate(DocumentEvent e) {
            panelChanged();
        }

        @Override
        public void tableChanged(TableModelEvent e) {
            panelChanged();
        }
    }

    /**
     * Class which binds list of {@link MeasurementCollection} 's to a
     * {@link JTable}. Get's the list of measurementCollections from the passed
     * sourceObject
     */
    class MeasurementTableBinder {

        private MeasurementTableModel tableModel;
        private List<MeasurementCollection> measurementCollections;
        private RowPersister persister;
        private Observation observation;
        private Map<MeasurementRow, MeasurementMapper> measurementMapping = new HashMap<>();

        public MeasurementTableBinder(Observation sourceObject, JTable destinationTable) {
            this.observation = sourceObject;
            this.measurementCollections = sourceObject.getMeasurementCollections();
            persister = new RowPersister();
            List<MeasurementRow> rows = new ArrayList<>();
            for (MeasurementCollection collection : measurementCollections) {
                for (Measurement measurement : collection.getMeasurements()) {
                    MeasurementRow measurementRow = new MeasurementRow(collection.getMeasurementType(), measurement.getValue());
                    rows.add(measurementRow);
                    addMeasurementMapping(measurement, measurementRow);
                }
            }
            tableModel = new MeasurementTableModel(rows);
            destinationTable.setModel(tableModel);
        }

        /**
         * Add a empty row to the JTable
         */
        public void addRow() {
            MeasurementRow row = new MeasurementRow();
            tableModel.addRow(row);
            row.addPropertyChangeListener(persister);
        }

        private void addMeasurementMapping(Measurement measurement, MeasurementRow measurementRow) {
            MeasurementMapper binder = new MeasurementMapper(measurement, measurementRow, observation);
            binder.bind();
            measurementMapping.put(measurementRow, binder);
        }

        /**
         * Remove a row at the passed index from the JTable and thus from the
         * Measurements in {@link Observation#measurementCollections}
         *
         * @param index
         */
        public void removeRow(int index) {
            MeasurementRow row = tableModel.removeRow(index);
            row.removePropertyChangeListener(persister);
            MeasurementMapper mapper = measurementMapping.get(row);
            if (mapper != null) {
                MeasurementCollection collection = ObservationUtil.getMeasurementCollection(observation.getMeasurementCollections(), row.getMeasurementType());
                collection.removeMeasurement(mapper.getMeasurement());
                measurementMapping.remove(row);
                if (collection.getMeasurements().isEmpty()) {
                    observation.removeMeasurementCollection(collection);
                }
            }
        }

        public boolean measurmentsValid() {
            boolean valid = true;
            for (MeasurementRow row : tableModel.getData()) {
                if (!isMeasurementRowvalid(row)) {
                    valid = false;
                }
            }
            return valid;
        }

        public boolean measurementTypesValid() {
            boolean valid = true;
            for (MeasurementRow row : tableModel.getData()) {
                if (!isMeasurementTypeValid(row)) {
                    valid = false;
                }
            }
            return valid;
        }

        public boolean measurementsValueValid() {
            boolean valid = true;
            for (MeasurementRow row : tableModel.getData()) {
                if (!isMeasurementValueValid(row)) {
                    valid = false;
                }
            }
            return valid;
        }

        private boolean isMeasurementRowvalid(MeasurementRow row) {
            return isMeasurementTypeValid(row) && isMeasurementValueValid(row);
        }

        /**
         * Check if the {@link Measurement#measurementType} is valid only valid
         * if not null
         *
         * @param measurementWrapper
         * @return
         */
        private boolean isMeasurementTypeValid(MeasurementRow measurementWrapper) {
            return measurementWrapper.getMeasurementType() != null;
        }

        /**
         * Check if the {@link Measurement#measurementValue} is valid Only valid
         * if not null
         *
         * @param measurementWrapper
         * @return
         */
        private boolean isMeasurementValueValid(MeasurementRow measurementWrapper) {
            return measurementWrapper.getValue() != null;
        }

        /**
         * Binder class between {@link Measurement} and {@link MeasurementRow}
         * Needs the source observation the passed measurement belongs to for
         * the event the {@link MeasurementType} changes
         */
        private class MeasurementMapper implements PropertyChangeListener {

            private Measurement measurement;
            private final MeasurementRow measurementRow;
            private final Observation observation;

            /**
             * Will create a new instance of MeasurementBinder
             *
             * @param measurement
             * @param measurementRow
             * @param observation
             */
            public MeasurementMapper(Measurement measurement, MeasurementRow measurementRow, Observation observation) {
                this.measurement = measurement;
                this.measurementRow = measurementRow;
                this.observation = observation;
            }

            public void bind() {
                measurement.addPropertyChangeListener(this);
                measurementRow.addPropertyChangeListener(this);
            }

            public Measurement getMeasurement() {
                return measurement;
            }

            @Override
            public void propertyChange(PropertyChangeEvent evt) {
                switch (evt.getPropertyName()) {
                    case Measurement.PROP_VALUE:
                        //the value if a measurement changed so sync to measurementrow
                        measurementRow.setValue((Double) evt.getNewValue());
                        break;
                    case MeasurementRow.PROP_MEASUREMENT_TYPE:
                        //The type of a row is changed needs to be removed from one collecion and added to another
                        MeasurementType oldType = (MeasurementType) evt.getOldValue();
                        MeasurementType newType = (MeasurementType) evt.getNewValue();
                        MeasurementCollection oldCollection = ObservationUtil.getMeasurementCollection(observation.getMeasurementCollections(), oldType);
                        oldCollection.removeMeasurement(measurement);
                        if (oldCollection.getMeasurements().isEmpty()) {
                            observation.removeMeasurementCollection(oldCollection);
                        }
                        MeasurementCollection newCollection = ObservationUtil.getMeasurementCollection(observation.getMeasurementCollections(), newType);
                        if (newCollection == null) {
                            newCollection = new MeasurementCollection(newType);
                            observation.addMeasurementCollection(newCollection);
                        }
                        newCollection.addMeasurement(measurement);
                        break;
                    case MeasurementRow.PROP_VALUE:
                        //Measurement row value changed need to update measurement
                        measurement.setValue((Double) evt.getNewValue());
                        break;
                }
            }
        }

        private class RowPersister implements PropertyChangeListener {

            @Override
            public void propertyChange(PropertyChangeEvent evt) {
                if (MeasurementRow.PROP_MEASUREMENT_TYPE.equals(evt.getPropertyName()) || MeasurementRow.PROP_VALUE.equals(evt.getPropertyName())) {
                    MeasurementRow row = (MeasurementRow) evt.getSource();
                    if (isMeasurementRowvalid(row)) {
                        persist(row);
                    }
                }
            }

            private void persist(MeasurementRow row) {
                MeasurementCollection toAddTo = ObservationUtil.getMeasurementCollection(measurementCollections, row.getMeasurementType());
                if (toAddTo == null) {
                    toAddTo = new MeasurementCollection(row.getMeasurementType());
                    observation.addMeasurementCollection(toAddTo);
                }
                Measurement measurement = new Measurement();
                measurement.setValue(row.getValue());
                Coordinate centre = new Coordinate(observation.getSourceGeometry().getCentroid().getX(), observation.getSourceGeometry().getCentroid().getY());
                measurement.setCoordinate(centre);
                toAddTo.addMeasurement(measurement);
                row.removePropertyChangeListener(this);

                addMeasurementMapping(measurement, row);
            }
        }

        @NbBundle.Messages({"MeasurementTableModel type column text=Type",
            "MeasurementTableModel value column text=Value",
            "MeasurementTableModel value unit column text=Unit",})
        class MeasurementTableModel extends AbstractTableModel {

            private final String COLUMN_TITLE_TYPE = Bundle.MeasurementTableModel_type_column_text();
            private final String COLUMN_TITLE_VALUE = Bundle.MeasurementTableModel_value_column_text();
            private final String COLUMN_TITLE_UOM = Bundle.MeasurementTableModel_value_unit_column_text();
            private List<MeasurementRow> measurementRows = new ArrayList();
            private final String[] COLUMN_NAMES = new String[]{COLUMN_TITLE_TYPE, COLUMN_TITLE_VALUE, COLUMN_TITLE_UOM};
            private final Class[] COLUMN_CLASS = new Class[]{MeasurementType.class, Double.class, UOM.class};

            public MeasurementTableModel(List<MeasurementRow> measurementRows) {
                this.measurementRows = measurementRows;
            }

            @Override
            public int getRowCount() {
                return measurementRows.size();
            }

            @Override
            public int getColumnCount() {
                return COLUMN_NAMES.length;
            }

            @Override
            public String getColumnName(int columnIndex) {
                return COLUMN_NAMES[columnIndex];
            }

            @Override
            public Class<?> getColumnClass(int columnIndex) {
                return COLUMN_CLASS[columnIndex];
            }

            @Override
            public boolean isCellEditable(int rowIndex, int columnIndex) {
                if (columnIndex == 2) {
                    return false;
                }
                return true;
            }

            @Override
            public Object getValueAt(int rowIndex, int columnIndex) {
                MeasurementRow wrapper = measurementRows.get(rowIndex);
                if (wrapper != null) {
                    switch (columnIndex) {
                        case 0:
                            return wrapper.getMeasurementType();
                        case 1:
                            return wrapper.getValue();
                        case 2:
                            UOM returnUom = UOM.NOT_SPECIFIED;
                            if(wrapper.getMeasurementType() != null){
                                returnUom = wrapper.getMeasurementType().getUnitOfMeasure();
                            }
                            return returnUom;
                    }
                }
                return null;
            }

            @Override
            public void setValueAt(Object aValue, int rowIndex, int columnIndex) {
                if (rowIndex < measurementRows.size()) {
                    MeasurementRow row = measurementRows.get(rowIndex);
                    switch (columnIndex) {
                        case 0:
                            MeasurementType measurementType = (MeasurementType) aValue;
                            row.setMeasurementType(measurementType);
                            fireTableCellUpdated(rowIndex, 2);
                            break;
                        case 1:
                            Double value = (Double) aValue;
                            row.setValue(value);
                            break;
                    }
                    fireTableCellUpdated(rowIndex, columnIndex);
                }
            }

            public void addRow(MeasurementRow measurementRow) {
                measurementRows.add(measurementRow);
                int pos = measurementRows.size();
                fireTableRowsInserted(pos, pos);

            }

            public MeasurementRow removeRow(int rowIndex) {
                fireTableRowsDeleted(rowIndex, rowIndex);
                MeasurementRow row = measurementRows.remove(rowIndex);
                return row;
            }

            public List<MeasurementRow> getData() {
                return measurementRows;
            }
        }

        /**
         * Row for the Measurement table model
         */
        private class MeasurementRow {

            private MeasurementType measurementType;
            public static final String PROP_MEASUREMENT_TYPE = "measurementType";
            private Double value;
            private static final String PROP_VALUE = "rowValue";
            private PropertyChangeSupport propertyChangeSupport;

            public MeasurementRow() {
            }

            public MeasurementRow(MeasurementType measurementType, Double value) {
                this.measurementType = measurementType;
                this.value = value;
            }

            /**
             * Get the type of measurement
             *
             * @return
             */
            public MeasurementType getMeasurementType() {
                return measurementType;
            }

            /**
             * Set the type of measurement
             *
             * @param measurementType
             */
            public void setMeasurementType(MeasurementType measurementType) {
                MeasurementType old = this.measurementType;
                this.measurementType = measurementType;
                getPropertyChangeSupport().firePropertyChange(PROP_MEASUREMENT_TYPE, old, measurementType);
            }

            public Double getValue() {
                return value;
            }

            public void setValue(Double value) {
                Double old = this.value;
                this.value = value;
                getPropertyChangeSupport().firePropertyChange(PROP_VALUE, old, value);
            }

            /**
             * Get the property change support
             *
             * @return
             */
            protected PropertyChangeSupport getPropertyChangeSupport() {
                if (propertyChangeSupport == null) {
                    propertyChangeSupport = new PropertyChangeSupport(this);
                }
                return propertyChangeSupport;
            }

            public void addPropertyChangeListener(PropertyChangeListener changeListener) {
                getPropertyChangeSupport().addPropertyChangeListener(changeListener);
            }

            public void addPropertyChangeListener(String prop, PropertyChangeListener changeListener) {
                getPropertyChangeSupport().addPropertyChangeListener(prop, changeListener);
            }

            public void removePropertyChangeListener(PropertyChangeListener changeListener) {
                getPropertyChangeSupport().removePropertyChangeListener(changeListener);
            }

            public void removePropertyChangeListener(String prop, PropertyChangeListener changeListener) {
                getPropertyChangeSupport().removePropertyChangeListener(prop, changeListener);
            }
        }
    }
}
