/**
 * Copyright (C) 2008-2012 AgroSense Foundation.
 *
 * AgroSense is free software: you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation, either version 3 of the License, or (at your option) any later
 * version.
 *
 * There are special exceptions to the terms and conditions of the GPLv3 as it
 * is applied to this software, see the FLOSS License Exception
 * <http://www.agrosense.eu/foss-exception.html>.
 *
 * AgroSense is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
 * A PARTICULAR PURPOSE. See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * AgroSense. If not, see <http://www.gnu.org/licenses/>.
 */
package nl.cloudfarming.client.sensor.visual;

import java.awt.Color;
import java.awt.Dimension;
import java.awt.Image;
import javax.swing.*;
import net.miginfocom.swing.MigLayout;
import nl.cloudfarming.client.sensor.model.MeasurementType;
import nl.cloudfarming.client.util.swing.FormRow;
import nl.cloudfarming.client.util.swing.MessagePanel;
import nl.cloudfarming.client.util.swing.button.ImageButton;
import org.jdesktop.swingx.JXDatePicker;
import org.openide.util.ImageUtilities;
import org.openide.util.NbBundle;

/**
 * Panel to enter new observation measurements
 *
 * @author Merijn Zengers <Merijn.Zengers@gmail.nl>
 */
@NbBundle.Messages({"ObservationVisualPanel name label text=Name",
    "ObservationVisualPanel sample date label text=Sample date",
    "ObservationVisualPanel add measurement button tooltip text=Add a measurement",
    "ObservationVisualPanel remove measurement button tooltip text=Remove selected measurement"})
public class ObservationVisualPanel extends JPanel {

    private JTextField nameField;
    private FormRow nameRow;
    private JXDatePicker sampleDatePicker;
    private FormRow dateRow;
    private JTable measurementsTable;
    private ImageButton addButton;
    private static final Image ADD_ICON = ImageUtilities.loadImage("nl/cloudfarming/client/sensor/visual/add.png");
    private static final Image ADD_ICON_SELECTED = ImageUtilities.loadImage("nl/cloudfarming/client/sensor/visual/add_sel.png");
    private static final Image ADD_ICON_DISABLED = ImageUtilities.loadImage("nl/cloudfarming/client/sensor/visual/add_dis.png");
    private static final String ADD_ICON_TOOLTIP_TEXT = Bundle.ObservationVisualPanel_add_measurement_button_tooltip_text();
    public static final String ADD_BUTTON_ACTION_COMMAND = "add";
    private ImageButton removeButton;
    private static final Image DELETE_ICON = ImageUtilities.loadImage("nl/cloudfarming/client/sensor/visual/delete.png");
    private static final Image DELETE_ICON_SELECTED = ImageUtilities.loadImage("nl/cloudfarming/client/sensor/visual/delete_sel16.png");
    private static final Image DELETE_ICON_DISABLED = ImageUtilities.loadImage("nl/cloudfarming/client/sensor/visual/delete_dis16.png");
    private final String DELETE_ICON_TOOLTIP_TEXT = Bundle.ObservationVisualPanel_remove_measurement_button_tooltip_text();
    public static final String DELETE_BUTTON_ACTION_COMMAND = "delete";
    private static final Color ICON_BACKGROUND_COLOR = Color.WHITE;
    private MessagePanel errorPanel;

    public ObservationVisualPanel() {
        init();
    }

    private void init() {
        this.setLayout(new MigLayout("wrap 1"));
        nameField = new JTextField();
        nameRow = new FormRow(true, Bundle.ObservationVisualPanel_name_label_text(), Bundle.ObservationVisualPanel_name_label_text(), nameField);
        sampleDatePicker = new JXDatePicker();
        dateRow = new FormRow(true, Bundle.ObservationVisualPanel_sample_date_label_text(), Bundle.ObservationVisualPanel_sample_date_label_text(), sampleDatePicker);

        add(nameRow);
        add(dateRow);

        JPanel buttonPanel = new JPanel(new MigLayout("wrap 2, insets 0 0 0 0"));
        buttonPanel.setBackground(ICON_BACKGROUND_COLOR);

        addButton = new ImageButton(ADD_ICON, ADD_ICON_SELECTED, ADD_ICON_DISABLED, ICON_BACKGROUND_COLOR, ADD_ICON_TOOLTIP_TEXT);
        addButton.setActionCommand(ADD_BUTTON_ACTION_COMMAND);
        buttonPanel.add(addButton, "gapleft 25");

        removeButton = new ImageButton(DELETE_ICON, DELETE_ICON_SELECTED, DELETE_ICON_DISABLED, ICON_BACKGROUND_COLOR, DELETE_ICON_TOOLTIP_TEXT);
        removeButton.setActionCommand(DELETE_BUTTON_ACTION_COMMAND);
        buttonPanel.add(removeButton);

        add(buttonPanel);

        measurementsTable = new JTable();
        JComboBox<MeasurementType> measurementComboBox = new JComboBox<>();
        measurementComboBox.setModel(new DefaultComboBoxModel<>(MeasurementType.values()));

        measurementsTable.setDefaultEditor(MeasurementType.class, new DefaultCellEditor(measurementComboBox));
        JScrollPane observationsPane = new JScrollPane(measurementsTable);
        observationsPane.setMinimumSize(new Dimension(dateRow.getWidth(), 0));
        add(observationsPane, "gapleft 25");

        errorPanel = new MessagePanel(MessagePanel.LARGE_ICON);
        add(errorPanel, "gapleft 25");

    }

    /**
     * Get the text field which holds the name
     *
     * @return
     */
    public JTextField getNameField() {
        return nameField;
    }

    /**
     * Get the row which displays the name
     *
     * @return
     */
    public FormRow getNameRow() {
        return nameRow;
    }

    /**
     * Get the date picker which holds the sample date
     *
     * @return
     */
    public JXDatePicker getSampleDatePicker() {
        return sampleDatePicker;
    }

    /**
     * Get the row which displays the sample date
     *
     * @return
     */
    public FormRow getSampleDateRow() {
        return dateRow;
    }

    /**
     * Get the table which displays the measurements
     *
     * @return
     */
    public JTable getMeasurementsTable() {
        return measurementsTable;
    }

    /**
     * Get the add button from this panel
     *
     * @return
     */
    public JButton getAddButton() {
        return addButton;
    }

    /**
     * Get the remove button from this panel
     *
     * @return
     */
    public JButton getRemoveButton() {
        return removeButton;
    }

    public MessagePanel getErrorPanel() {
        return errorPanel;
    }

    
}
