/**
 * Copyright (C) 2008-2012 AgroSense Foundation.
 *
 * AgroSense is free software: you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation, either version 3 of the License, or (at your option) any later
 * version.
 *
 * There are special exceptions to the terms and conditions of the GPLv3 as it
 * is applied to this software, see the FLOSS License Exception
 * <http://www.agrosense.eu/foss-exception.html>.
 *
 * AgroSense is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
 * A PARTICULAR PURPOSE. See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * AgroSense. If not, see <http://www.gnu.org/licenses/>.
 */
package nl.cloudfarming.client.sensor.visual.util;

import com.vividsolutions.jts.geom.Coordinate;
import com.vividsolutions.jts.geom.Geometry;
import eu.agrosense.client.grid.Grid;
import eu.agrosense.client.grid.GridProvider;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import nl.cloudfarming.client.farm.model.FarmManager;
import nl.cloudfarming.client.model.EnterpriseIdProvider;
import nl.cloudfarming.client.sensor.model.Measurement;
import nl.cloudfarming.client.sensor.model.MeasurementCollection;
import nl.cloudfarming.client.sensor.model.MeasurementType;
import nl.cloudfarming.client.sensor.model.Observation;
import org.netbeans.api.project.Project;
import org.openide.util.Lookup;

/**
 * Utility class for {@link Observation}
 *
 * @author Merijn Zengers <Merijn.Zengers@gmail.nl>
 */
public class ObservationUtil {

    /**
     * Creates a grid for a list of {@link Measurement} 's. Will check if there
     * are {@link Measurement} with the same {@link Measurement#coordinate} if
     * this is the case will calculate an average of the values and places this
     * average in the grid. Returns null if there are no {@link Measurement} 's
     * in the observation of the passed type
     *
     * @param observation
     * @param newMeasurement
     * @return the created grid or null is the passed list of measurements is
     * empty
     */
    public static Grid createGrid(Observation observation, List<Measurement> measurements) {
        if (!measurements.isEmpty()) {
            GridProvider gridProvider = Lookup.getDefault().lookup(GridProvider.class);
            Geometry envelope = observation.getSourceGeometry().getEnvelope();
            Coordinate min = envelope.getCoordinates()[0];
            Coordinate max = envelope.getCoordinates()[2];
            FarmManager farmManager = Lookup.getDefault().lookup(FarmManager.class);
            Project project = farmManager.findFarmProject(observation.getSource());
            Grid grid = gridProvider.getDefaultGridSection(project.getLookup().lookup(EnterpriseIdProvider.class), min, max);
            Map<Coordinate, List<Measurement>> coordinateMap = getMeasuremtCoordinateMap(measurements);
            for (Coordinate coordinate : coordinateMap.keySet()) {
                Double average = getAverageValue(coordinateMap.get(coordinate));
                grid.setRasterValue(coordinate, average.intValue());
            }

            return grid;
        }

        return null;
    }

    /**
     * Updates the provided grid with the newly provided {@link Measurement}
     * Will search the provided list of measurements for measurements with the
     * same coordinate. If found will take an average of the values on the same
     * coordinate
     *
     * @param measurementsInGrid the measurements currently in the grid
     * @param grid the grid to update
     * @param newMeasurement the measurement to add to the grid
     */
    public static void addMeasurementToGrid(List<Measurement> measurementsInGrid, Grid grid, Measurement newMeasurement) {
        Double value = newMeasurement.getValue();
        if (!measurementsInGrid.isEmpty()) {
            List<Measurement> sameCoordinate = getMeasurementsForCoordinate(measurementsInGrid, newMeasurement.getCoordinate());
            sameCoordinate.add(newMeasurement);
            if (!sameCoordinate.isEmpty()) {
                value = getAverageValue(sameCoordinate);
            }
        }
        grid.setRasterValue(newMeasurement.getCoordinate(), value.intValue());

    }

    /**
     * Updates the provided grid. Removes the provided {@link Measurement} from
     * it Will search the provided list of {@link Measurement} 's for
     * measurements with the same coordinate If found will take an average of
     * the values of the remaining measurements and set it in the grid
     *
     * @param measurementsInGrid the measurements currently in the grid
     * @param grid the grid to update
     * @param oldMeasurement the measurement to remove from the grid
     */
    public static void removeMeasurementFromGrid(List<Measurement> measurementsInGrid, Grid grid, Measurement oldMeasurement) {
        Double value = Double.valueOf(Grid.EMPTY_VALUE);
        if (!measurementsInGrid.isEmpty()) {
            List<Measurement> sameCoordinate = getMeasurementsForCoordinate(measurementsInGrid, oldMeasurement.getCoordinate());
            sameCoordinate.remove(oldMeasurement);
            if (!sameCoordinate.isEmpty()) {
                value = getAverageValue(sameCoordinate);
            }
        }
        grid.setRasterValue(oldMeasurement.getCoordinate(), value.intValue());
    }

    /**
     * Get the average value of the passed measurements
     *
     * @param measurements
     * @return
     */
    public static Double getAverageValue(List<Measurement> measurements) {
        assert !measurements.isEmpty() : "cannot calculate a avereage from an empty list of measurements";
        Double average = Double.valueOf(0);
        for (Measurement measurement : measurements) {
            Double value = measurement.getValue();
            assert value != null : "Cannot calculate average without a value present in the measurment";
            average += value;
        }
        return average / measurements.size();
    }

    /**
     * Get the {@link Measurement} 's for a coordinate
     *
     * @param measurements the measurements to search
     * @param coordinate the coordinate to search for
     * @return
     */
    public static List<Measurement> getMeasurementsForCoordinate(List<Measurement> measurements, Coordinate coordinate) {
        List<Measurement> returnList = new ArrayList<>();
        for (Measurement measurement : measurements) {
            if (coordinate.equals(measurement.getCoordinate())) {
                returnList.add(measurement);
            }
        }
        return returnList;
    }

    /**
     * Gets a map of coordinates mapping to measurments with this coordinate
     *
     * @param measurements
     * @return
     */
    public static Map<Coordinate, List<Measurement>> getMeasuremtCoordinateMap(List<Measurement> measurements) {
        Map<Coordinate, List<Measurement>> coordinateMap = new HashMap<>();
        for (Measurement measurement : measurements) {
            Coordinate coordinate = measurement.getCoordinate();
            assert coordinate != null : "Cannot create coordinate map without a coordinate";
            if (coordinateMap.containsKey(coordinate)) {
                coordinateMap.get(coordinate).add(measurement);
            } else {
                List<Measurement> coordinateMeasurements = new ArrayList<>();
                coordinateMeasurements.add(measurement);
                coordinateMap.put(coordinate, coordinateMeasurements);
            }
        }
        return coordinateMap;
    }

    /**
     * Get a {@link MeasurementCollection} from a list of measurementCollections
     * searching for the provided {@link MeasurementType}
     *
     * @param collections
     * @param measurementType
     * @return
     */
    public static MeasurementCollection getMeasurementCollection(List<MeasurementCollection> collections, MeasurementType measurementType) {
        for (MeasurementCollection collection : collections) {
            if (collection.getMeasurementType().equals(measurementType)) {
                return collection;
            }
        }
        return null;
    }
}
